# Authors: marcusm117
# License: Apache 2.0
"""This Module contains a class KripkeStruct, which is a graph implementation of Kripke Structures.

We plan to support other Transition Systems as well in the future.

"""

# Standard Libraries
from typing import List, Dict, Set, Tuple
from collections import defaultdict, UserDict
from copy import deepcopy


class KripkeStructError(Exception):
    """Exceptions raised by methods related to class KripkeStruct."""


class KripkeStruct:
    """Class that implements a Kripke Structure for Model Checking.

    An instance of thie class can be created from a JSON file or from scratch.
    For more information about Kripke Structures, see:
    https://en.wikipedia.org/wiki/Kripke_structure_(model_checking).

    Attributes:
        _atoms (tuple): Atoms, can only be reset before any state is created
        _atoms_set (set): Atoms, a duplicate of "_atoms", only for efficiency
        _states (_UniqueValueDict): States, Key is the state name, Value is the state label
        _starts (set): Start States
        _trans (defaultdict): Transitions, Key is the source state, Value is a set of target states
        _trans_inverted (defaultdict): Inverted Transitions, Key is the target state, Value is a set of source states

    """

    def __init__(self, model_json=None):
        self._atoms = ()
        self._atoms_set = set()
        self._states = self._KripkeStateDict()
        self._starts = set()
        self._trans = defaultdict(set)
        self._trans_inverted = defaultdict(set)

        if model_json is not None:
            self.set_atoms(model_json["Atoms"])
            self.add_states(model_json["States"])
            self.set_starts(model_json["Starts"])
            self.add_trans(model_json["Trans"])

    class _KripkeStateDict(UserDict):
        def __init__(self, atoms_num: int = 0):
            self.atoms_num = atoms_num
            super().__init__()

        def __setitem__(self, key, value):
            # if the value is already assigned to the key, do nothing
            if (key in self) and (self.__getitem__(key) == value):
                return

            # if the value is NOT an int, raise an error
            if not isinstance(value, int):
                raise KripkeStructError("The a State Label should be an int, when assigning with []")

            # if the value exceeds 2^n, where n is the number of atoms, raise an error
            if value >= 2**self.atoms_num:
                raise KripkeStructError("The value should be less than 2^n, where n is the number of atoms")

            # if the value is already assigned to a different key, raise an error
            # no matter if the key exists or not
            if value in self.values():
                raise KripkeStructError("Can't assign an existing State Label to a different State Name")

            super().__setitem__(key, value)

    def __str__(self) -> str:
        return (
            f"Atoms: {self._atoms}\n"
            + f"States: {dict(self._states)}\n"
            + f"Starts: {self._starts}\n"
            + f"Trans: {dict(self._trans)}"
        )

    def set_atoms(self, atoms: List[str]) -> None:
        """Set Atoms of the Kripke Structure.

        Args:
            atoms: a list of strings representing the Atoms

        Raises:
            KripkeStructError: if any state exists, can't reset atoms

        Note:
            The parameter "atoms" should be a list to stay compatible with the JSON format.
            However, internally, the field "_atoms" is stored as a tuple for efficiency.

        """
        # if any state exists, can't reset atoms
        if self._states:
            raise KripkeStructError("Can't reset Atoms after States are Created")
        self._atoms = tuple(atoms)
        self._atoms_set = set(atoms)
        self._states.atoms_num = len(atoms)

    def get_atoms(self) -> Tuple[str]:
        """Get Atoms of the Kripke Structure.

        Returns:
            a tuple of strings representing the Atoms

        """
        return self._atoms

    def add_state(self, state: str, label: List[str]) -> None:
        """Add a State to the Kripke Structure.

        The State Label is a list of strings, where each string is an Atom.
        For example, if the atom is ("a", "b", "c"),
        then the label can be ["a"], ["a", "b", "c"], or [].

        Args:
            state: a string representing the State Name
            label: a list of strings representing the State Label

        Raises:
            KripkeStructError: if the State Name or State Label exists, can't add again
            KrinkeStructError: if the State Label contains a Non-Eixsting Atom, can't add it

        Note:
            The parameter "label" should be a list to stay compatible with the JSON format.
            However, internally, the field "_states" is stored as a dict[str, int] for efficiency.

        """
        # if the state exists, can't add again
        if state in self._states:
            raise KripkeStructError("Can't add an Existing State Name again")

        # convert the label to a set, for efficiency,
        # since we'll check if each atom in "self._atoms" is in "label" later
        label_set = set(label)

        # if the State Label contains a Non-Eixsting Atom, can't add it
        for atom in label_set:
            if atom not in self._atoms_set:
                raise KripkeStructError("Can't add a State Label with a Non-Existing Atom")

        # convert the label to its binary representation, to save memory
        # for example, if the atoms are ("a", "b", "c", "d"), and the label is ["a", "c"],
        # then the binary representation is 0b1010, which is the int 10 in decimal
        label_binary = 0b0
        for atom in self._atoms:
            # if an Atom is in "label", set the corresponding bit to 1
            if atom in label_set:
                label_binary |= 0b1
            # left shift by 1 bit
            label_binary <<= 1

        # correct the extra left shift
        label_binary >>= 1

        # if the state label exists, can't add again
        if label_binary in self._states.values():
            raise KripkeStructError("Can't add an Existing State Label again")

        self._states[state] = label_binary

    def add_states(self, states: Dict[str, List[str]]) -> None:
        """Add multiple States to the Kripke Structure.

        Args:
            states: a dict, Key is the State Name, Value is the State Label

        Raises:
            KripkeStructError: if a State Name or a State Label exists, can't add again

        """
        for state, label in states.items():
            self.add_state(state, label)

    def get_states(self) -> Dict[str, int]:
        """Get States of the Kripke Structure.

        Returns:
            a dict, Key is the State Name, Value is the State Label

        Note:
            The State Label is store in the binary representation,
            instead of the list of strings.

        """
        return dict(self._states)

    def get_state_names(self) -> Set[str]:
        """Get just State Names of the Kripke Structure.

        Returns:
            a set of strings representing the State Names

        """
        return set(self._states.keys())

    def set_label_of_state(self, state: str, label_set: Set[str]) -> None:
        """Given an existing State Name, (re)set the State Label.

        Args:
            state: a string representing the State Name
            label_set: a set of strings representing the State Label

        Raises:
            KripkeStructError: if the State Name doesn't exist, can't (re)set the State Label
            KripkeStructError: if the Label has been assigned to a differnt State Name, can't assign it

        Note:
            The parameter "label_set" should be a set, NOT a list, for efficiency.

        """
        # if the State Name doesn't exist, can't set the Label of it
        if state not in self._states:
            raise KripkeStructError("Can't set the Label of a Non-Existing State")

        # if the State Label contains a Non-Eixsting Atom, can't set it
        for atom in label_set:
            if atom not in self._atoms_set:
                raise KripkeStructError("Can't set a State Label with a Non-Existing Atom")

        # convert the label to its binary representation
        label_binary = 0b0
        for atom in self._atoms:
            # if an Atom is in "label", set the corresponding bit to 1
            if atom in label_set:
                label_binary |= 0b1
            # left shift by 1 bit
            label_binary <<= 1

        # correct the extra left shift
        label_binary >>= 1

        # if the Label is already assigned to the State Name, do nothing
        if self._states[state] == label_binary:
            return

        # if the Label is assigned to a differnt State Name, can't assign it
        if label_binary in self._states.values():
            raise KripkeStructError("Can't assign an Existing State Label to a Different State Name")

        self._states[state] = label_binary

    def get_label_of_state(self, state: str) -> Set[str]:
        """Given a State Name, get the corresponding State Label.

        For example, if atoms is ("a", "b", "c", "d"),
        the State Name is "s1", and the State Label is 0b1010,
        then get_label_of_state("s1") will return the set {"a", "c"}.

        Args:
            state: a string representing the State Name

        Returns:
            a set of strings representing the State Label, instead of the binary form

        Raises:
            KripkeStructError: if the State Name doesn't exist, can't get the Label of it

        """
        # if the State Name doesn't exist, can't get the Label of it
        if state not in self._states:
            raise KripkeStructError("Can't get the Label of a Non-Existing State")

        # get the binary form of the Label, then convert it to a set of atoms
        label = self._states[state]
        length = len(self._atoms)
        label_set = set()

        # traverse the binary form of the Label
        for i in range(length):
            # if the i-th bit is 1, then the (length - 1 - i)-th atom is in the set
            if label & 1 << i:
                label_set.add(self._atoms[length - 1 - i])

        return label_set

    def remove_state(self, state: str) -> None:
        """Remove a State from the Kripke Structure.

        Args:
            state: a string representing the State Name

        Raises:
            KripkeStructError: if the State Name doesn't exist, can't remove it

        """
        # if the State Name doesn't exist, can't remove it
        if state not in self._states:
            raise KripkeStructError("Can't remove a Non-Existing State")
        self._states.pop(state)

        # removing a state will remove all related transitions
        if state in self._trans:
            next_states = self._trans.pop(state)
            for next_state in next_states:
                self._trans_inverted[next_state].remove(state)
        if state in self._trans_inverted:
            prev_states = self._trans_inverted.pop(state)
            for prev_state in prev_states:
                self._trans[prev_state].remove(state)

    def remove_states(self, states: List[str]) -> None:
        """Remove multiple States from the Kripke Structure.

        Args:
            states: a list of strings representing the State Names

        """
        for state in states:
            self.remove_state(state)

    def set_starts(self, starts: List[str]) -> None:
        """Set Start States of the Kripke Structure.

        Args:
            starts: a set of strings representing the Start States

        Raises:
            KripkeStructError: if a state doesn't exist, can't set it as a Start State

        Note:
            The parameter "starts" should be a list to stay compatible with the JSON format.
            However, internally, the field "_starts" is stored as a set for efficiency.

        """
        for start in starts:
            # if a state doesn't exist, can't set it as a Start State
            if start not in self._states:
                raise KripkeStructError("Can't set a Non-Existing State as Start State")
        self._starts = set(starts)

    def get_starts(self) -> Set[str]:
        """Get Start States of the Kripke Structure.

        Returns:
            a set of strings representing the Start States

        """
        return self._starts

    def add_trans(self, trans: Dict[str, Set[str]]) -> None:
        """Add multiple Transitions to the Kripke Structure.

        Args:
            trans: a dict, Key is the Source State Name, Value is a list of Target State Names

        Raises:
            KripkeStructError: if Source or Target State doesn't exist, can't add transition

        Note:
            The parameter "trans" should be a dict whose value is a list, to stay compatible with the JSON format.
            However, internally, the field "_trans" is stored as a dict whose value is a set for efficiency.
            For more information about list v.s. set, see:
            https://stackoverflow.com/questions/2831212/python-sets-vs-lists
        """
        for state, next_states in trans.items():
            # if source state doesn't exist, can't add transition
            if state not in self._states:
                raise KripkeStructError("Can't add Transition from a Non-Existing Source State")

            for next_state in next_states:
                # if target state doesn't exist, can't add transition
                if next_state not in self._states:
                    raise KripkeStructError("Can't add Transition to a Non-Existing Target State")
                # adding a Transition will also update the Inverted Transitions,
                # which will be used to remove related Transitions when revmoing a State
                self._trans[state].add(next_state)
                self._trans_inverted[next_state].add(state)

    def get_trans(self) -> defaultdict[str, Set[str]]:
        """Get Transitions of the Kripke Structure.

        Returns:
            a defaultdict, Key is the Source State Name, Value is a set of Target State Names

        """
        return self._trans

    def get_trans_inverted(self) -> defaultdict[str, Set[str]]:
        """Get Inverted Transitions of the Kripke Structure.

        Returns:
            a defaultdict, Key is the Target State Name, Value is a set of Source State Names

        """
        return self._trans_inverted

    def remove_trans(self, trans: Dict[str, List[str]]) -> None:
        """Remove multiple Transitions from the Kripke Structure.

        Args:
            trans: a dict, Key is the Source State Name, Value is a list of Target State Names

        Raises:
            KripkeStructError: if a Transition doesn't exist, can't remove it

        Note:
            The parameter "trans" should be a dict whose value is a list for efficiency,
            since it will only be iterated once.
            However, internally, the field "_trans" is stored as a dict whose value is a set for efficiency.

        """
        for state, next_states in trans.items():
            for next_state in next_states:
                # if a Transition doesn't exist, can't remove it
                if (state not in self._trans) or (next_state not in self._trans[state]):
                    raise KripkeStructError("Can't remove a Non-Existing Transition")
                # removing a Transition will also update the Inverted Transitions
                self._trans[state].remove(next_state)
                self._trans_inverted[next_state].remove(state)

    def reverse_all_trans(self) -> None:
        """Reverse all Transitions in the Kripke Structure.

        This funciton is destructive!!!
        It will NOT create a new copy of the current Kripke Structure,
        but will directly modify the current Kripke Structure.

        """
        self._trans, self._trans_inverted = self._trans_inverted, self._trans

    def get_SCCs(self) -> List[Set[str]]:
        """Get all Strongly Connected Components (SCCs) in the Kripke Structure.

        Returns:
            a list of sets, where each set contains the State Names in a SCC

        """
        # we use Kosaraju's Algorithm to find SCCs
        # first, we run DFS on the original graph, store the order of finishing states in "order_stack"
        order_stack = []
        visited = set()

        def DFS(state, visited, order_stack):
            visited.add(state)
            successors = self._trans[state]
            for successor in successors:
                if successor not in visited:
                    DFS(successor, visited, order_stack)
            # add to order_stack when exiting DFS
            order_stack.append(state)

        for state in self._states:
            if state not in visited:
                DFS(state, visited, order_stack)

        # second, we create the reversed graph of the Kripke Structure self
        # since we need to modify the reversed graph, we create a deepcopy of self
        reversed_graph = deepcopy(self)
        reversed_graph.reverse_all_trans()

        # third we run multiple rounds of DFS on the reversed graph until all states are visited
        # when we finish a round of DFS, we get a SCC
        # then we add the SCC to the set, and remove all states in the SCC from the reversed graph
        SCCs = set()
        while reversed_graph.get_states():
            # we run DFS on the reversed graph, following the order of the "order_stack"
            tmp_stack = [order_stack.pop()]
            visited = set()
            while tmp_stack:
                current_state = tmp_stack.pop()
                visited.add(current_state)

                # remove the visited state from the order_stack
                if current_state in order_stack:
                    order_stack.remove(current_state)

                # add un-visited successors to the stack
                successors = reversed_graph.get_trans()[current_state]
                for successor in successors:
                    if successor not in visited:
                        tmp_stack.append(successor)

            # now all states in the set "visited" are in the same SCC
            # so we add it to the list of SCCs, and remove all states in it from the reversed graph
            SCCs.add(frozenset(visited))
            reversed_graph.remove_states(list(visited))

        return SCCs



# Authors: marcusm117
# License: Apache 2.0

# Standard Libraries
from copy import deepcopy

# External Libraries
import pytest

# Internal Modules to be tested
from mctk import KripkeStruct, KripkeStructError
from mctk import SAT_atom, NOT, AND, OR, IMPLIES, IFF, EX, AX, EU, EF, AG, EG, AF, AU


ks_json = {
    "Atoms": ["a", "b", "c", "d"],
    "States": {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
        "s5": ["b"],
        "s6": ["c"],
        "s7": ["d"],
    },
    "Starts": ["s1"],
    "Trans": {
        's1': ['s2'],
        's2': ['s3', 's4'],
        's3': ['s4'],
        's4': ['s7'],
        's5': ['s6'],
        's6': ['s7', 's5'],
        's7': ['s5'],
    },
}
ks = KripkeStruct(ks_json)


def test_ESMC_SAT_atom():
    sat_states = SAT_atom(ks, "a")
    assert sat_states == {"s1", "s2"}

    sat_states = SAT_atom(ks, "b")
    assert sat_states == {"s2", "s3", "s4", "s5"}

    sat_states = SAT_atom(ks, "c")
    assert sat_states == {"s3", "s4", "s6"}

    sat_states = SAT_atom(ks, "d")
    assert sat_states == {"s4", "s7"}

    # if the atomic property is not in the Kripke Structure, can't check it
    with pytest.raises(KripkeStructError) as error_info:
        sat_states = SAT_atom(ks, "e")
    assert str(error_info.value) == "Can't check on an atom that's not in the Kripke Structure"

    sat_states = SAT_atom(ks, "True")
    assert sat_states == set(ks.get_states().keys())

    sat_states = SAT_atom(ks, "False")
    assert sat_states == set()


def test_ESMC_NOT():
    sat_states = NOT(ks, SAT_atom(ks, "a"))
    assert sat_states == set(ks.get_states().keys()) - {"s1", "s2"}

    sat_states = NOT(ks, SAT_atom(ks, "b"))
    assert sat_states == set(ks.get_states().keys()) - {"s2", "s3", "s4", "s5"}

    sat_states = NOT(ks, SAT_atom(ks, "True"))
    assert sat_states == set()

    sat_states = NOT(ks, SAT_atom(ks, "False"))
    assert sat_states == set(ks.get_states().keys())


def test_ESMC_AND():
    sat_states = AND(SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    assert sat_states == {"s2"}

    sat_states = AND(SAT_atom(ks, "a"), SAT_atom(ks, "c"))
    assert sat_states == set()

    sat_states = AND(SAT_atom(ks, "b"), SAT_atom(ks, "c"))
    assert sat_states == {"s3", "s4"}


def test_ESMC_OR():
    sat_states = OR(SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5"}

    sat_states = OR(SAT_atom(ks, "a"), SAT_atom(ks, "c"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s6"}


def test_ESMC_IMPLIES():
    sat_states = IMPLIES(ks, SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    assert sat_states == {"s2", "s3", "s4", "s5", "s6", "s7"}

    sat_states = IMPLIES(ks, SAT_atom(ks, "a"), SAT_atom(ks, "c"))
    assert sat_states == {"s3", "s4", "s5", "s6", "s7"}

    sat_states = IMPLIES(ks, SAT_atom(ks, "b"), SAT_atom(ks, "c"))
    assert sat_states == {"s1", "s3", "s4", "s6", "s7"}


def test_ESMC_IFF():
    sat_states = IFF(ks, SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    assert sat_states == {"s2", "s6", "s7"}

    sat_states = IFF(ks, SAT_atom(ks, "a"), SAT_atom(ks, "c"))
    assert sat_states == {"s5", "s7"}

    sat_states = IFF(ks, SAT_atom(ks, "b"), SAT_atom(ks, "c"))
    assert sat_states == {"s1", "s3", "s4", "s7"}


def test_ESMC_EX():
    sat_states = EX(ks, SAT_atom(ks, "a"))
    assert sat_states == {"s1"}

    sat_states = EX(ks, SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s6", "s7"}

    sat_states = EX(ks, SAT_atom(ks, "c"))
    assert sat_states == {"s2", "s3", "s5"}

    sat_states = EX(ks, SAT_atom(ks, "d"))
    assert sat_states == {"s2", "s3", "s4", "s6"}


def test_ESMC_AX():
    sat_states = AX(ks, SAT_atom(ks, "a"))
    assert sat_states == {"s1"}

    sat_states = AX(ks, SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s7"}

    sat_states = AX(ks, SAT_atom(ks, "c"))
    assert sat_states == {"s2", "s3", "s5"}

    sat_states = AX(ks, SAT_atom(ks, "d"))
    assert sat_states == {"s3", "s4"}


def test_ESMC_EU():
    sat_states = EU(ks, SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5"}

    sat_states = EU(ks, SAT_atom(ks, "a"), SAT_atom(ks, "c"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s6"}

    sat_states = EU(ks, SAT_atom(ks, "b"), SAT_atom(ks, "c"))
    assert sat_states == {"s2", "s3", "s4", "s5", "s6"}


def test_ESMC_EF():
    sat_states = EF(ks, SAT_atom(ks, "a"))
    assert sat_states == {"s1", "s2"}

    sat_states = EF(ks, SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5", "s6", "s7"}

    sat_states = EF(ks, SAT_atom(ks, "c"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5", "s6", "s7"}

    sat_states = EF(ks, SAT_atom(ks, "d"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5", "s6", "s7"}


def test_ESMC_AG():
    sat_states = AG(ks, SAT_atom(ks, "a"))
    assert sat_states == set()

    sat_states = AG(ks, SAT_atom(ks, "b"))
    assert sat_states == set()

    sat_states = AG(ks, SAT_atom(ks, "c"))
    assert sat_states == set()

    sat_states = AG(ks, SAT_atom(ks, "d"))
    assert sat_states == set()


def test_ESMC_EG():
    # change s5's label to {"b", "d"}
    # change s6's label to {"c", "d"}
    tmp_ks = deepcopy(ks)
    tmp_ks.set_label_of_state("s5", {"b", "d"})
    tmp_ks.set_label_of_state("s6", {"c", "d"})

    sat_states = EG(tmp_ks, SAT_atom(tmp_ks, "a"))
    assert sat_states == set()

    sat_states = EG(tmp_ks, SAT_atom(tmp_ks, "b"))
    assert sat_states == set()

    sat_states = EG(tmp_ks, SAT_atom(tmp_ks, "c"))
    assert sat_states == set()

    sat_states = EG(tmp_ks, SAT_atom(tmp_ks, "d"))
    assert sat_states == {"s4", "s5", "s6", "s7"}

    tmp_ks.remove_trans({"s4": ["s7"]})
    sat_states = EG(tmp_ks, SAT_atom(tmp_ks, "d"))
    assert sat_states == {"s5", "s6", "s7"}


def test_ESMC_AF():
    # change s7 to "", which is empty set
    tmp_ks = deepcopy(ks)
    tmp_ks.set_label_of_state("s7", set())

    sat_states = AF(tmp_ks, SAT_atom(tmp_ks, "a"))
    assert sat_states == {"s1", "s2"}

    sat_states = AF(tmp_ks, SAT_atom(tmp_ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5", "s6", "s7"}

    sat_states = AF(tmp_ks, SAT_atom(tmp_ks, "c"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5", "s6", "s7"}

    sat_states = AF(tmp_ks, SAT_atom(tmp_ks, "d"))
    assert sat_states == {"s1", "s2", "s3", "s4"}


def test_ESMC_AU():
    sat_states = AU(ks, SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5"}

    sat_states = AU(ks, SAT_atom(ks, "a"), SAT_atom(ks, "c"))
    assert sat_states == {"s1", "s2", "s3", "s4", "s6"}

    sat_states = AU(ks, SAT_atom(ks, "b"), SAT_atom(ks, "c"))
    assert sat_states == {"s2", "s3", "s4", "s5", "s6"}


# Integrated Tests
def test_ESMC_composite_CTL_formula():
    sat_states = EF(ks, SAT_atom(ks, "a") & SAT_atom(ks, "b"))
    assert sat_states == {"s1", "s2"}

    sat_states = AU(ks, NOT(ks, SAT_atom(ks, "a")), SAT_atom(ks, "c"))
    assert sat_states == {"s3", "s4", "s5", "s6", "s7"}

    sat_states = EX(ks, AF(ks, SAT_atom(ks, "b")))
    assert sat_states == {"s1", "s2", "s3", "s4", "s5", "s6", "s7"}

    sat_states = AU(ks, EX(ks, SAT_atom(ks, "b")), SAT_atom(ks, "c"))
    print(sat_states)
    assert sat_states == {"s1", "s2", "s3", "s4", "s6"}


# Integrated Tests: examples in documentations
def test_ESMC_examples():
    # create a Kripke Structure from scratch
    ks = KripkeStruct()

    # set 2 Atomic Propositions in this Kripke Structure
    ks.set_atoms(["p", "q"])

    # add 2 states to the Kripke Structure
    # a State Name is represented by a string, it must be unique
    # a State Label is represented by a binary number, it must be unique
    # for example, 0b10 means the state has the Atoms "p" but not "q"
    ks.add_state("s0", ["p"])
    ks.add_state("s1", ["q"])

    # set the Start States of the Kripke Structure
    # there can be multiple Start States
    ks.set_starts(["s0"])

    # add 2 Transitions to the Kripke Structure
    # a Transition is represented by a key-value pair
    # where key the Source State Name and value is a list of Destination State Names
    ks.add_trans({"s0": ["s1"], "s1": ["s0"]})

    # check if the Kripke Structure satisfies the CTL formula: EX p
    # SAT_atom(ks, "p") returns a set of states that satisfy the Atomic Proposition p
    # EX returns a set of states that satisfy the CTL formula EX p
    sat_states = EX(ks, SAT_atom(ks, "p"))

    # the result should be {"s1"}
    # since the start state "s0" is not in sat_states, ks doesn't satisfy the CTL formula
    assert sat_states == {"s1"}

    # check if the Kripke Structure satisfies the CTL formula: E p U q
    # SAT_atom(ks, "p") returns a set of states that satisfy the Atomic Proposition p
    # SAT_atom(ks, "q") returns a set of states that satisfy the Atomic Proposition q
    # EU returns a set of states that satisfy the CTL formula E p U q
    sat_states = EU(ks, SAT_atom(ks, "p"), SAT_atom(ks, "q"))

    # the result should be {"s0", "s1"}
    # since the start state "s0" is in sat_states, ks satisfies the CTL formula
    assert sat_states == {"s0", "s1"}

    # composite CTL formula is supported as the following
    # check if the Kripke Structure satisfies the CTL formula: EX (p AND EX q)
    sat_states = EX(ks, AND(SAT_atom(ks, "p"), EX(ks, SAT_atom(ks, "q"))))

    # the result should be {"s1"}
    # since the start state "s0" is not in sat_states, ks doesn't satisfy the CTL formula
    assert sat_states == {"s1"}

    # check if the Kripke Structure satisfies the CTL formula: EG (A p U (NOT q))
    sat_states = EG(ks, AU(ks, SAT_atom(ks, "p"), NOT(ks, SAT_atom(ks, "q"))))

    # the result should be set(), empty set
    # since the start state "s0" is not in sat_states, ks doesn't satisfy the CTL formula
    assert sat_states == set()

    # check CTL formula on a Complex Kripke Structure
    ks_json = {
        "Atoms": ["a", "b", "c", "d"],
        "States": {
            "s1": ["a"],
            "s2": ["a", "b"],
            "s3": ["b", "c"],
            "s4": ["b", "c", "d"],
            "s5": ["b"],
            "s6": ["c"],
            "s7": ["d"],
        },
        "Starts": ["s1"],
        "Trans": {
            's1': ['s2'],
            's2': ['s3', 's4'],
            's3': ['s4'],
            's4': ['s7'],
            's5': ['s6'],
            's6': ['s7', 's5'],
            's7': ['s5'],
        },
    }
    ks = KripkeStruct(ks_json)

    # check if the Kripke Structure satisfies the CTL formula: EX a
    sat_states = EX(ks, SAT_atom(ks, "a"))

    # the result should be {"s1"}
    # since the start state "s1" is in sat_states, ks satisfies the CTL formula
    assert sat_states == {"s1"}

    # check if the Kripke Structure satisfies the CTL formula: E a U b
    sat_states = EU(ks, SAT_atom(ks, "a"), SAT_atom(ks, "b"))

    # the result should be {'s1', 's2', 's3', 's4', 's5'}
    # since the start state "s1" is in sat_states, ks satisfies the CTL formula
    assert sat_states == {'s1', 's2', 's3', 's4', 's5'}

    # check if the Kripke Structure satisfies the CTL formula: EG a
    sat_states = EG(ks, SAT_atom(ks, "a"))

    # the result should be set()
    # since the start state "s1" is not in sat_states, ks doesn't satisfy the CTL formula
    assert sat_states == set()



# Authors: marcusm117
# License: Apache 2.0
"""This Module contains functions for Explicit-State Model Checking CTL properties on the class KripkeStruct.

We plan to support Symbolic Model Checking and Bounded Model Checking in the future.

Examples:
    >>> from mctk import *
    >>> ks_json = {
    ...     "Atoms": ["a", "b", "c", "d"],
    ...     "States": {
    ...         "s1": 0b1000,  # s1 has labels "a"
    ...         "s2": 0b1100,  # s2 has labels "a", "b"
    ...         "s3": 0b0110,  # s3 has labels "b", "c"
    ...         "s4": 0b0111,  # s4 has labels "b", "c", "d"
    ...         "s5": 0b0100,  # s5 has label "b"
    ...         "s6": 0b0010,  # s6 has label "c"
    ...         "s7": 0b0001,  # s7 has label "d"
    ...     },
    ...     "Starts": ["s1"],
    ...     "Trans": {
    ...         's1': ['s2'],
    ...         's2': ['s3', 's4'],
    ...         's3': ['s4'],
    ...         's4': ['s7'],
    ...         's5': ['s6'],
    ...         's6': ['s7', 's5'],
    ...         's7': ['s5'],
    ...     },
    ... }
    >>> ks = KripkeStruct(ks_json)
    >>> sat_states = SAT_atom(ks, "a")
    >>> print(sat_states)
    {'s1', 's2'}
    >>> sat_states = AND(SAT_atom(ks, "b"), SAT_atom(ks, "c"))
    >>> print(sat_states)
    {'s3', 's4'}
    >>> sat_states = EX(ks, SAT_atom(ks, "a"))
    >>> print(sat_states)
    {'s2'}
    >>> sat_states = EU(ks, SAT_atom(ks, "a"), SAT_atom(ks, "b"))
    >>> print(sat_states)
    {'s1', 's2', 's3', 's4', 's5'}
    >>> sat_states = EG(tmp_ks, SAT_atom(tmp_ks, "a"))
    >>> print(sat_states)
    set()

"""


# Standard Libraries
from copy import deepcopy
from typing import Set

# Internal Modules
from .models import KripkeStruct, KripkeStructError


def SAT_atom(ks: KripkeStruct, atomic_property: str) -> Set[str]:
    """Evaluate an atomic propositional property on a Kripke Structure.

    Return a set of states where the atomic propositional property is satisfied.

    Args:
        ks: a Kripke Structure
        atomic_property: an Atomic Propositional Property represented as a string

    Returns:
        a set of states where the Atomic Propositional Property is satisfied

    Raises:
        KripkeStructError: if the Atomic Property is not in the Kripke Structure

    """
    sat_states = set()
    state_names = ks.get_state_names()

    if atomic_property == "True":
        return state_names
    if atomic_property == "False":
        return set()
    if atomic_property not in ks.get_atoms():
        raise KripkeStructError("Can't check on an atom that's not in the Kripke Structure")

    for state in state_names:
        # get the Label for the State Name, which is a set of atoms
        label = ks.get_label_of_state(state)

        # if the atomic property is in the Label, then this state satisfies the atomic property
        if atomic_property in label:
            sat_states.add(state)

    return sat_states


def NOT(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the NOT operator on a CTL property.

    Return a set of states where the CTL formula "NOT property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "NOT property" is satisfied

    """
    # complement of the set
    return ks.get_state_names() - property1


def AND(property1: Set[str], property2: Set[str]) -> Set[str]:
    """Implement the AND operator on two CTL properties.

    Return a set of states where the CTL formula "property1 AND property2" is satisfied.

    Args:
        ks: a Kripke Structure
        property1: a CTL property represented as a set of states that satisfy the property
        property2: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "property1 AND property2" is satisfied

    """
    # intersection of the two sets
    return property1 & property2


def OR(property1: Set[str], property2: Set[str]) -> Set[str]:
    """Implement the OR operator on two CTL properties.

    Return a set of states where the CTL formula "property1 OR property2" is satisfied.

    Args:
        ks: a Kripke Structure
        property1: a CTL property represented as a set of states that satisfy the property
        property2: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "property1 OR property2" is satisfied

    """
    # union of the two sets
    return property1 | property2


def IMPLIES(ks: KripkeStruct, property1: Set[str], property2: Set[str]) -> Set[str]:
    """Implement the IMPLIES operator on two CTL properties.

    Return a set of states where the CTL formula "property1 IMPLIES property2" is satisfied.

    Args:
        ks: a Kripke Structure
        property1: a CTL property represented as a set of states that satisfy the property
        property2: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "property1 IMPLIES property2" is satisfied

    """
    # complement of property1, then union property2
    # p IMPLIES q = NOT(p) OR q
    return NOT(ks, property1) | property2


def IFF(ks: KripkeStruct, property1: Set[str], property2: Set[str]) -> Set[str]:
    """Implement the IFF operator on two CTL properties.

    Return a set of states where the CTL formula "property1 IFF property2" is satisfied.

    Args:
        ks: a Kripke Structure
        property1: a CTL property represented as a set of states that satisfy the property
        property2: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "property1 IFF property2" is satisfied

    """
    # (p IFF q) = (p IMPLIES q) AND (q IMPLIES p)
    return IMPLIES(ks, property1, property2) & IMPLIES(ks, property2, property1)


def EX(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the EX operator on a CTL property.

    Return a set of states where the CTL formula "EX property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "EX property" is satisfied

    """
    sat_states = set()

    for state in ks.get_states():
        # get the successors of the state
        successors = ks.get_trans()[state]

        for successor in successors:
            # test if the successors satisfy the property
            if successor in property1:
                sat_states.add(state)

    return sat_states


def AX(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the AX operator on a CTL property.

    Return a set of states where the CTL formula "AX property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "AX property" is satisfied

    """
    # AX p = NOT (EX NOT(p))
    return NOT(ks, EX(ks, NOT(ks, property1)))


def EU(ks: KripkeStruct, property1: Set[str], property2: Set[str]) -> Set[str]:
    """Implement the EU operator on two CTL properties.

    Return a set of states where the CTL formula "E property1 U property2" is satisfied.

    Args:
        ks: a Kripke Structure
        property1: a CTL property represented as a set of states that satisfy the property
        property2: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "E property1 U property2" is satisfied

    """
    # if a state satisfies property2, then it satisfies "E property1 U property2" by definition
    # therefore we intialize set_stats to be property2
    sat_states = property2
    size = len(sat_states)

    # we keep adding states that can reach "sat_states" in 1 step, and also satisfy property1
    # until we reach a fixed point
    # i.e. no new states are added to "sat_states"
    # i.e. size of "sat_states" does not change
    while True:
        # get the predecessors of all states in "sat_states"
        predecessors = set()
        for state in sat_states:
            predecessors = predecessors | ks.get_trans_inverted()[state]

        # add states that can reach "sat_states" in 1 step, and also satisfy property1
        sat_states = sat_states | (property1 & predecessors)

        # if no new states are added, then we have reached a fixed point
        new_size = len(sat_states)
        if new_size == size:
            break
        size = new_size

    return sat_states


def EF(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the EF operator on a CTL property.

    Return a set of states where the CTL formula "EF property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "EF property" is satisfied

    """
    # EF p = E true U p
    return EU(ks, ks.get_state_names(), property1)


def AG(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the AG operator on a CTL property.

    Return a set of states where the CTL formula "AG property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "AG property" is satisfied

    """
    # AG p = NOT (EF NOT(p))
    return NOT(ks, EF(ks, NOT(ks, property1)))


def EG(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the EG operator on a CTL property.

    Return a set of states where the CTL formula "EG property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where "EG property" is satisfied

    """
    sat_states = set()

    # creat a sub-graph of the original Kripke Structure ks
    # where we remove all states that do not satisfy property
    # since we need to modify the reversed graph, we create a deepcopy of self
    sub_graph = deepcopy(ks)
    for state in ks.get_states():
        if state not in property1:
            sub_graph.remove_state(state)

    # compute all Strongly Connected Components (SCCs) of the sub-graph
    # SCCs is a set of SCCs, where each SCC is a set of states
    SCCs = sub_graph.get_SCCs()

    # add all non-trivial SCCs to sat_states
    for SCC in SCCs:
        if len(SCC) > 1:
            sat_states = sat_states | SCC

    # if there are no non-trivial SCCs, then no states satisfy EG property
    if not sat_states:
        return sat_states

    # otherwise, we keep adding states that can reach "sat_states" in 1 step
    # until we reach a fixed point
    # i.e. no new states are added to "sat_states"
    # i.e. size of "sat_states" does not change
    size = len(sat_states)
    while True:
        # get the predecessors of all states in "sat_states"
        predecessors = set()
        for state in sat_states:
            predecessors = predecessors | sub_graph.get_trans_inverted()[state]

        # add states that can reach "sat_states" in 1 step
        sat_states = sat_states | predecessors

        # if no new states are added, then we have reached a fixed point
        new_size = len(sat_states)
        if new_size == size:
            break
        size = new_size

    return sat_states


def AF(ks: KripkeStruct, property1: Set[str]) -> Set[str]:
    """Implement the AF operator on a CTL property.

    Return a set of states where the CTL formula "AF property" is satisfied.

    Args:
        ks: a Kripke Structure
        property: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "AF property" is satisfied

    """
    # AF p = NOT (EG NOT(p))
    return NOT(ks, EG(ks, NOT(ks, property1)))


def AU(ks: KripkeStruct, property1: Set[str], property2: Set[str]) -> Set[str]:
    """Implement the AU operator on two CTL properties.

    Return a set of states where the CTL formula "A property1 U property2" is satisfied.

    Args:
        ks: a Kripke Structure
        property1: a CTL property represented as a set of states that satisfy the property
        property2: a CTL property represented as a set of states that satisfy the property

    Returns:
        a set of states where the CTL formula "A property1 U property2" is satisfied

    """
    # A p1 U p2 = NOT (E NOT(p2) U (NOT(p1) AND NOT(p2))) AND NOT (EG NOT(p2))
    #           = NOT (E NOT(p2) U (NOT(p1) AND NOT(p2))) AND AF p2
    return NOT(ks, EU(ks, NOT(ks, property2), NOT(ks, property1) & NOT(ks, property2))) & NOT(
        ks, EG(ks, NOT(ks, property2))
    )



# Authors: marcusm117
# License: Apache 2.0

# Standard Libraries
from collections import defaultdict

# External Libraries
import pytest

# Internal Modules to be tested
from mctk import KripkeStruct, KripkeStructError


# Tests for KripkeStruct()
def test_KS_default_init():
    ks = KripkeStruct()
    assert ks._atoms == ()
    assert ks._atoms_set == set()
    assert ks._states == {}
    assert ks._starts == set()
    assert ks._trans == defaultdict(set)
    assert ks._trans_inverted == defaultdict(set)


# Tests for KripkeStruct(ks_json)
def test_KS_json_init():
    ks_json = {
        "Atoms": ["a", "b", "c", "d"],
        "States": {"s1": ["a"], "s2": ["b"], "s3": ["c", "d"], "s4": []},
        "Starts": ["s1"],
        "Trans": {"s1": ["s2"], "s2": ["s3", "s4"], "s3": ["s4"]},
    }
    ks = KripkeStruct(ks_json)

    assert ks._atoms == ("a", "b", "c", "d")
    assert ks._atoms_set == {"a", "b", "c", "d"}
    assert ks._states == {"s1": 0b1000, "s2": 0b0100, "s3": 0b0011, "s4": 0b0000}
    assert ks._starts == {"s1"}
    assert ks._trans == {"s1": {"s2"}, "s2": {"s3", "s4"}, "s3": {"s4"}}
    assert ks._trans_inverted == {"s2": {"s1"}, "s3": {"s2"}, "s4": {"s2", "s3"}}


# Tests for ks.__str__()
def test_KS_str_rep():
    ks_json = {
        "Atoms": ["a", "b", "c", "d"],
        "States": {"s1": ["a"], "s2": ["b"], "s3": ["c", "d"], "s4": []},
        "Starts": ["s1"],
        "Trans": {"s1": ["s2"], "s2": ["s3"], "s3": ["s4"]},
    }
    ks = KripkeStruct(ks_json)
    assert str(ks) == (
        "Atoms: ('a', 'b', 'c', 'd')\n"
        + "States: {'s1': 8, 's2': 4, 's3': 3, 's4': 0}\n"
        + "Starts: {'s1'}\n"
        + "Trans: {'s1': {'s2'}, 's2': {'s3'}, 's3': {'s4'}}"
    )


def test_KS_set_atoms():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    assert ks._atoms == ("a", "b", "c", "d")
    assert ks._atoms_set == {"a", "b", "c", "d"}

    # if any state exists, can't reset atoms
    with pytest.raises(KripkeStructError) as error_info:
        ks.add_state("s1", ["a"])
        atoms = ("a",)
        ks.set_atoms(atoms)
    assert str(error_info.value) == "Can't reset Atoms after States are Created"


def test_KS_get_atoms():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    assert ks.get_atoms() == ("a", "b", "c", "d")


def test_KS_add_state():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    ks.add_state("s1", ["a"])
    ks.add_state("s2", ["a", "b"])
    assert ks._states == {"s1": 0b1000, "s2": 0b1100}

    # if the state name exists, can't add again
    with pytest.raises(KripkeStructError) as error_info:
        ks.add_state("s1", ["a", "b", "c", "d"])
    assert str(error_info.value) == "Can't add an Existing State Name again"

    # if the state label exists, can't add again
    with pytest.raises(KripkeStructError) as error_info:
        ks.add_state("s3", ["a"])
    assert str(error_info.value) == "Can't add an Existing State Label again"

    # if the State Label contains a Non-Eixsting Atom, can't add it
    with pytest.raises(KripkeStructError) as error_info:
        ks.add_state("s3", ["a", "b", "c", "d", "e"])
    assert str(error_info.value) == "Can't add a State Label with a Non-Existing Atom"


def test_KS_add_states():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
        "s5": ["b"],
        "s6": ["c"],
        "s7": ["d"],
    }
    ks.add_states(states)
    assert ks._states == {
        "s1": 0b1000,  # s1 has labels "a"
        "s2": 0b1100,  # s2 has labels "a", "b"
        "s3": 0b0110,  # s3 has labels "b", "c"
        "s4": 0b0111,  # s4 has labels "b", "c", "d"
        "s5": 0b0100,  # s5 has label "b"
        "s6": 0b0010,  # s6 has label "c"
        "s7": 0b0001,  # s7 has label "d"
    }


def test_KS_get_states():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    ks.add_state("s1", ["a"])
    assert ks.get_states() == {"s1": 0b1000}


def test_KS_get_state_names():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    ks.add_state("s1", ["a"])
    assert ks.get_state_names() == {"s1"}


def test_KS_set_label_of_state():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    ks.add_state("s1", ["a"])
    ks.set_label_of_state("s1", {"c", "d"})
    assert ks.get_states() == {"s1": 0b0011}

    # if the State Name doesn't exist, can't set the Label of it
    with pytest.raises(KripkeStructError) as error_info:
        ks.set_label_of_state("s2", {"a", "b", "c", "d"})
    assert str(error_info.value) == "Can't set the Label of a Non-Existing State"

    # if the Label has been assigned to a differnt State Name, can't assign it
    with pytest.raises(KripkeStructError) as error_info:
        ks.add_state("s2", ["a", "b", "c", "d"])
        ks.set_label_of_state("s2", {"c", "d"})
    assert str(error_info.value) == "Can't assign an Existing State Label to a Different State Name"

    # if the State Label contains a Non-Eixsting Atom, can't set it
    with pytest.raises(KripkeStructError) as error_info:
        ks.set_label_of_state("s2", {"a", "b", "c", "d", "e"})
    assert str(error_info.value) == "Can't set a State Label with a Non-Existing Atom"

    # assigning the Label to its corresponding State Name, wont' have any effect
    ks.set_label_of_state("s2", {"a", "b", "c", "d"})
    assert ks.get_states() == {"s1": 0b0011, "s2": 0b1111}

    # if the Label is assigned to a differnt State Name, can't assign it using []
    with pytest.raises(KripkeStructError) as error_info:
        ks._states["s2"] = 0b0011
    assert str(error_info.value) == "Can't assign an existing State Label to a different State Name"

    # if the Label is NOT an int, can't assign it using []
    with pytest.raises(KripkeStructError) as error_info:
        ks._states["s2"] = {"a"}
    assert str(error_info.value) == "The a State Label should be an int, when assigning with []"

    # if the label value exceeds 2^n, where n is the number of atoms, can't assign it using []
    with pytest.raises(KripkeStructError) as error_info:
        ks._states["s2"] = 0b100000
    assert str(error_info.value) == "The value should be less than 2^n, where n is the number of atoms"

    # assigning the Label to its corresponding State Name using [], wont' have any effect
    ks._states["s2"] = 0b1111
    assert ks.get_states() == {"s1": 0b0011, "s2": 0b1111}


def test_KS_get_label_of_state():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    ks.add_state("s1", ["a", "d"])
    assert ks.get_label_of_state("s1") == {"a", "d"}

    # if the state doesn't exist, can't get the Label of it
    with pytest.raises(KripkeStructError) as error_info:
        assert ks.get_label_of_state("s2")
    assert str(error_info.value) == "Can't get the Label of a Non-Existing State"


def test_KS_remove_state():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
        "s5": ["b"],
        "s6": ["c"],
        "s7": ["d"],
    }
    ks.add_states(states)
    ks.remove_state("s7")
    assert ks.get_states() == {
        "s1": 0b1000,  # s1 has labels "a"
        "s2": 0b1100,  # s2 has labels "a", "b"
        "s3": 0b0110,  # s3 has labels "b", "c"
        "s4": 0b0111,  # s4 has labels "b", "c", "d"
        "s5": 0b0100,  # s5 has label "b"
        "s6": 0b0010,  # s6 has label "c"
    }

    # if the State Name doesn't exist, can't remove it
    with pytest.raises(KripkeStructError) as error_info:
        ks.remove_state("s8")
    assert str(error_info.value) == "Can't remove a Non-Existing State"


def test_KS_remove_states():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
        "s5": ["b"],
        "s6": ["c"],
        "s7": ["d"],
    }
    ks.add_states(states)
    ks.remove_states(["s6", "s7"])
    assert ks.get_states() == {
        "s1": 0b1000,  # s1 has labels "a"
        "s2": 0b1100,  # s2 has labels "a", "b"
        "s3": 0b0110,  # s3 has labels "b", "c"
        "s4": 0b0111,  # s4 has labels "b", "c", "d"
        "s5": 0b0100,  # s5 has label "b"
    }


def test_KS_set_starts():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)

    starts = ["s1", "s4"]
    ks.set_starts(starts)
    assert ks._starts == {"s1", "s4"}

    # resetting the start states is allowed
    starts = ["s1"]
    ks.set_starts(starts)
    assert ks._starts == {"s1"}

    # if state doesn't exist, can't set as start state
    with pytest.raises(KripkeStructError) as error_info:
        ks.set_starts(["s5"])
    assert str(error_info.value) == "Can't set a Non-Existing State as Start State"


def test_KS_get_starts():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)

    starts = ["s1", "s4"]
    ks.set_starts(starts)
    assert ks.get_starts() == {"s1", "s4"}


def test_KS_add_trans():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)
    ks.set_starts(["s1"])

    trans = {
        "s1": ["s2"],  # a -> ab
        "s2": ["s3", "s4"],  # ab -> bc
        "s3": ["s4", "s1"],  # bc -> bcd, a
        "s4": ["s2"],  # bcd -> ab
    }
    ks.add_trans(trans)
    assert ks._trans == {"s1": {"s2"}, "s2": {"s3", "s4"}, "s3": {"s4", "s1"}, "s4": {"s2"}}

    # if source state doesn't exist, can't add transition from it
    with pytest.raises(KripkeStructError) as error_info:
        trans = {"s7": ["s1"]}
        ks.add_trans(trans)
    assert str(error_info.value) == "Can't add Transition from a Non-Existing Source State"

    # if target state doesn't exist, can't add transition to it
    with pytest.raises(KripkeStructError) as error_info:
        trans = {"s1": ["s7"]}
        ks.add_trans(trans)
    assert str(error_info.value) == "Can't add Transition to a Non-Existing Target State"


def test_KS_get_trans():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)
    ks.set_starts(["s1"])
    trans = {
        "s1": ["s2"],  # a -> ab
        "s2": ["s3", "s4"],  # ab -> bc
        "s3": ["s4", "s1"],  # bc -> bcd, a
        "s4": ["s2"],  # bcd -> ab
    }
    ks.add_trans(trans)
    assert ks.get_trans() == {"s1": {"s2"}, "s2": {"s3", "s4"}, "s3": {"s4", "s1"}, "s4": {"s2"}}


def test_KS_get_trans_inverted():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)
    ks.set_starts(["s1"])
    trans = {
        "s1": ["s2"],  # a -> ab
        "s2": ["s3", "s4"],  # ab -> bc
        "s3": ["s4", "s1"],  # bc -> bcd, a
        "s4": ["s2"],  # bcd -> ab
    }
    ks.add_trans(trans)
    assert ks.get_trans_inverted() == {"s1": {"s3"}, "s2": {"s1", "s4"}, "s3": {"s2"}, "s4": {"s2", "s3"}}


def test_KS_remove_trans():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)
    ks.set_starts(["s1"])
    trans = {
        "s1": ["s2"],  # a -> ab
        "s2": ["s3", "s4"],  # ab -> bc
        "s3": ["s4", "s1"],  # bc -> bcd, a
        "s4": ["s2"],  # bcd -> ab
    }
    ks.add_trans(trans)

    trans = {"s2": ["s4"], "s4": ["s2"]}
    ks.remove_trans(trans)
    assert ks.get_trans() == {"s1": {"s2"}, "s2": {"s3"}, "s3": {"s4", "s1"}, "s4": set()}

    # if a Transition doesn't exist, can't remove it
    with pytest.raises(KripkeStructError) as error_info:
        ks.remove_trans({"s2": ["s4"]})
    assert str(error_info.value) == "Can't remove a Non-Existing Transition"


def test_KS_reverse_all_trans():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)
    ks.set_starts(["s1"])
    trans = {
        "s1": ["s2"],  # a -> ab
        "s2": ["s3", "s4"],  # ab -> bc
        "s3": ["s4", "s1"],  # bc -> bcd, a
        "s4": ["s2"],  # bcd -> ab
    }
    ks.add_trans(trans)

    ks.reverse_all_trans()
    assert ks.get_trans() == {"s1": {"s3"}, "s2": {"s1", "s4"}, "s3": {"s2"}, "s4": {"s2", "s3"}}


def test_KS_get_SCCs():
    ks_json = {
        "Atoms": ["a", "b", "c", "d"],
        "States": {
            "s1": ["a"],
            "s2": ["a", "b"],
            "s3": ["b", "c"],
            "s4": ["b", "c", "d"],
            "s5": ["b"],
            "s6": ["c"],
            "s7": ["d"],
        },
        "Starts": ["s1"],
        "Trans": {
            's1': ['s2'],
            's2': ['s3', 's4'],
            's3': ['s4'],
            's4': ['s7'],
            's5': ['s6'],
            's6': ['s7', 's5'],
            's7': ['s5'],
        },
    }
    ks = KripkeStruct(ks_json)
    SCC_1 = frozenset({"s1"})
    SCC_2 = frozenset({"s2"})
    SCC_3 = frozenset({"s3"})
    SCC_4 = frozenset({"s4"})
    SCC_5 = frozenset({"s5", "s6", "s7"})
    assert ks.get_SCCs() == {SCC_1, SCC_2, SCC_3, SCC_4, SCC_5}

    ks.add_trans({"s4": ["s2"]})
    SCC_6 = frozenset({"s2", "s3", "s4"})
    assert ks.get_SCCs() == {SCC_1, SCC_6, SCC_5}

    ks.remove_states(["s3", "s4", "s6"])
    SCC_7 = frozenset({"s5"})
    SCC_8 = frozenset({"s7"})
    assert ks.get_SCCs() == {SCC_1, SCC_2, SCC_7, SCC_8}


# Integration Tests: removing state will remove related transitions
def test_KS_remove_states_will_remove_related_trans():
    ks = KripkeStruct()
    atoms = ["a", "b", "c", "d"]
    ks.set_atoms(atoms)
    states = {
        "s1": ["a"],
        "s2": ["a", "b"],
        "s3": ["b", "c"],
        "s4": ["b", "c", "d"],
    }
    ks.add_states(states)
    ks.set_starts(["s1"])
    trans = {
        "s1": ["s2"],  # a -> ab
        "s2": ["s3", "s4"],  # ab -> bc
        "s3": ["s4", "s1"],  # bc -> bcd, a
        "s4": ["s2"],  # bcd -> ab
    }
    ks.add_trans(trans)

    ks.remove_state("s2")
    assert ks.get_trans() == {'s1': set(), 's3': {'s4', 's1'}, 's4': set()}
    assert ks.get_trans_inverted() == {'s3': set(), 's4': {'s3'}, 's1': {'s3'}}



from .models import KripkeStruct, KripkeStructError
from .checking import SAT_atom, NOT, AND, OR, IMPLIES, IFF, EX, AX, EF, AF, EG, AG, EU, AU

__version__ = "0.1.1"
__all__ = [
    "KripkeStruct",
    "KripkeStructError",
    "SAT_atom",
    "NOT",
    "AND",
    "OR",
    "IMPLIES",
    "IFF",
    "EX",
    "AX",
    "EF",
    "AF",
    "EG",
    "AG",
    "EU",
    "AU",
]
